//-------------------------------------------------------------------------------------------------
/// \file	BasicHwShader.cpp
/// \date	6-11-04
/// \author	Rob Bateman
/// \brief	This impliments possibly one of the most basic Hardware shaders available. It simply
///			aims to recreate the basic OpenGL material properties, ie ambient, diffuse, specular
/// 		emission and shininess.
///				This time we now have a an additional Drag and Drop behaviour class to handle
/// 		connections between this node and other maya nodes.
///
/// \note	To use this shader, you must connect the outColor attribute of the shader to the
/// 		hardwareShader input on a lambert, phong, blinn, phongE or anisotropic material. You
/// 		must also ensure that you are viewing the shader in textured mode, ie press 6 when
/// 		in the viewport.
///
/// \todo	Before this plugin is fully ready for production use, we still have to handle the
/// 		following :
///
/// 		1. Textures as yet are unsupported. ie, if you connect the color attribute of the
/// 		   hardware shader to a file texture, we will not get the texturing as yet...
///
//-------------------------------------------------------------------------------------------------

#include "BasicHwShader.h"
#include <maya/MGlobal.h>
#include <maya/MPlugArray.h>
#include <maya/MImage.h>

const MTypeId BasicHwShader::m_TypeId( 0x70030 );
const MString BasicHwShader::m_TypeName( "basicHwShader" );


/// The ambient Colour of the material
MObject  BasicHwShader::m_aAmbientColor;

/// The diffuse colour of the material
MObject  BasicHwShader::m_aDiffuseColor;

/// The specular colour for the material
MObject  BasicHwShader::m_aSpecularColor;

/// The emission colour for the material
MObject  BasicHwShader::m_aEmissionColor;

/// The shininess of the material
MObject  BasicHwShader::m_aShininess;

//
void BasicHwShader::postConstructor()
{
	// tell maya that this node is not safe for multiple processors.
	setMPSafe(false);
}

//
void * BasicHwShader::creator()
{
    return new BasicHwShader();
}

//
MStatus BasicHwShader::initialize()
{
	// we will use a numeric attribute function set to add all of
	// the attributes to the node description.
    MFnNumericAttribute nAttr;

    // create the ambient colour attribute
    m_aAmbientColor = nAttr.createColor( "ambientColor", "ambc");
    nAttr.setStorable(true);
    nAttr.setKeyable(true);
	nAttr.setDefault(0.3f, 0.3f, 0.3f);

    // create the diffuse colour attribute
	m_aDiffuseColor =  nAttr.createColor( "color", "c");
	nAttr.setStorable(true);
	nAttr.setKeyable(true);
	nAttr.setDefault(0.7f, 0.7f, 0.7f);

    // create the specular colour attribute
    m_aSpecularColor = nAttr.createColor( "specularColor", "sc");
	nAttr.setStorable(true);
	nAttr.setKeyable(true);
	nAttr.setDefault(1.0f, 1.0f, 1.0f);

    // create the emission colour attribute
    m_aEmissionColor =  nAttr.createColor( "incandescence", "ic");
	nAttr.setStorable(true);
	nAttr.setKeyable(true);
	nAttr.setDefault(0.0f, 0.0f, 0.0f);

    // create the materials shininess attribute
    m_aShininess = nAttr.create( "cosinePower", "cp",
						   MFnNumericData::kFloat,
						   10.0f);
	nAttr.setStorable(true);
	nAttr.setKeyable(true);
	nAttr.setMin(0.0f);
	nAttr.setMax(90.0f);

	// add the various attributes to the node
	addAttribute(m_aAmbientColor);
	addAttribute(m_aDiffuseColor);
	addAttribute(m_aSpecularColor);
	addAttribute(m_aEmissionColor);
	addAttribute(m_aShininess);

	// set the attribute dependencies. Probably only require the diffuse to outColor... oh well ;)
	attributeAffects(m_aAmbientColor,  outColor);
	attributeAffects(m_aDiffuseColor,  outColor);
	attributeAffects(m_aSpecularColor, outColor);
	attributeAffects(m_aEmissionColor, outColor);
	attributeAffects(m_aShininess,     outColor);

	// done
    return MS::kSuccess;
}

// don't need any colours
int	BasicHwShader::colorsPerVertex() {
	return 0;
}

// just the normals please ;)
int	BasicHwShader::normalsPerVertex() {
	return 1;
}

// only one set of uv's thank you...
int	BasicHwShader::texCoordsPerVertex() {
	return 1;
}

// nah, no transparency please...
bool BasicHwShader::hasTransparency() {
	return false;
}

// can't be bothered with vertex id's thank you....
bool BasicHwShader::provideVertexIds() {
	return false;
}

// utility func
void BasicHwShader::GetFloat(MObject oAttr, float & val) const
{
	// create a plug to the attribute on this node
	MPlug plug(thisMObject(), oAttr);

	// use the plug to get the float attribute
	plug.getValue(val);
}

// utility func to retrieve the colour for the specified attr
void BasicHwShader::GetColour(MObject oAttr, float val[]) const
{
	// create a plug to the attribute on this node
	MPlug	plug(thisMObject(), oAttr);

	// use the plug to get the compound attribute
	MObject object;
	plug.getValue(object);

	// attach a numeric attribute function set to the color
	MFnNumericData fn(object);

	// get the data from the color and wang in val
	fn.getData(val[0], val[1], val[2]);
}

/// \brief	This function returns true if the specified size is a power of two.
/// \param	i	-	the length value
/// \return	true if OK
///
bool IsPowerOfTwo(unsigned int i) {
	if( i == 2    ||
		i == 4    ||
		i == 8    ||
		i == 16   ||
		i == 32   ||
		i == 64   ||
		i == 128  ||
		i == 256  ||
		i == 512  ||
		i == 1024 ||
		i == 2048 ||
		i == 4096 )
	{
		return true;
	}
	return false;
}

/// \brief	An internal utility function to simplfy the process of getting texture data.
///
/// \param	oAttr	-	one of the static MObject attribute references on the node
/// \param	val		-	the returned openGL texture object
///
bool BasicHwShader::GetGlTexture(MObject oAttr, unsigned int& val) const {

	// create a plug to the attribute on this node
	MPlug	plug(thisMObject(), oAttr);

	// get the
	MPlugArray plugs;
	plug.connectedTo(plugs,true,false);

	val = 0;

	for(unsigned int i=0;i<plugs.length();++i)
	{
		if(plugs[i].node().hasFn(MFn::kFileTexture))
		{
			// get the image data from the file texture node
			MImage image;
			if( MS::kSuccess != image.readFromTextureNode(plugs[i].node()) )
			{
				// failed
				continue;
			}

			// get the image dimensions
			unsigned int w,h;
			image.getSize(w,h);

			// check to make sure that the image is of power 2
			if( !IsPowerOfTwo(w) ||
				!IsPowerOfTwo(h) )
			{
				// if not resize it to a default. You could probably use the width
				// and height to calculate the best size for the texture. Instead however
				// i'm just using a default of 256x256
				//
				w = h = 256;
				image.resize(w,h,false);
			}

			// generate a texture object number
			glGenTextures(1,&val);
			glPixelStorei (GL_UNPACK_ALIGNMENT, 1);

			// set the current texture to the one just created
			glBindTexture (GL_TEXTURE_2D, val);

			// repeat the texture in both directions
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

			// use bi-linear filtering on the texture
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

			// load the texture data into the graphics hardware.
			glTexImage2D (GL_TEXTURE_2D,0,GL_RGBA,w,h,0,GL_RGBA,GL_UNSIGNED_BYTE,image.pixels());

			// release the image data
			image.release();

			return true;
		}
	}
	return false;
}

/// \brief	This function is used to bind any textures before rendering of the surfaces begins.
/// \param	path	-	the DAG path to the shape to render.
/// \return	MS::kSuccess or MS::kFailure.
///
MStatus	BasicHwShader::glBind( const MDagPath& path ) {

	// save the current OpenGL and vertex array settings
    glPushAttrib(GL_ALL_ATTRIB_BITS);
	glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);

	// first attempt to get a file texture from the color attribute
	if( GetGlTexture(m_aDiffuseColor,m_TextureObject) )
	{
		// suceeded, so enable texturing and bind our texture
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D,m_TextureObject);
	}

	return MS::kSuccess;
}

/// \brief	This function is used to unbind any textures after the surfaces have been rendered
/// \param	path	-	the DAG path to the shape to render.
/// \return	MS::kSuccess or MS::kFailure.
///
MStatus	BasicHwShader::glUnbind( const MDagPath& path ) {

	// if the texture was loaded last time, make sure we delete the GL texture object
	if(m_TextureObject) {
		glDisable(GL_TEXTURE_2D);
		glDeleteTextures(1,&m_TextureObject);
		m_TextureObject=0;
	}

	// restore the previous OpenGL settings
	glPopClientAttrib();
	glPopAttrib();
	return MS::kSuccess;
}


// draw me some geometry using vertex arrays...
MStatus	BasicHwShader::glGeometry( const MDagPath&,
                                int prim,
								unsigned int writable,
								int indexCount,
								const unsigned int * indexArray,
								int vertexCount,
								const int * vertexIDs,
								const float * vertexArray,
								int normalCount,
								const float ** normalArrays,
								int colorCount,
								const float ** colorArrays,
								int texCoordCount,
								const float ** texCoordArrays)
{
	// these vars will hold the values stored in the nodes attributes
    float Ambient[] = {0.3f, 0.3f, 0.3f,1.0f};
    float Diffuse[] = {0.7f, 0.7f, 0.7f,1.0f};
    float Specular[] = {1.0f, 1.0f, 1.0f,1.0f};
    float Emission[] = {0.0f, 0.0f, 0.0f,0.0f};
    float Shininess = 10.0f;

	MGlobal::displayInfo("rob");

	// get the colours and shininess from the node
	GetColour(m_aAmbientColor, Ambient);
	GetColour(m_aSpecularColor, Specular);
	GetColour(m_aEmissionColor, Emission);
	GetFloat(m_aShininess, Shininess);

	// doing something slightly different with the diffuse colour now. If a texture is applied,
	// then the colour always gets evaluated as 0,0,0. Therefore, hack it ;)
	if(m_TextureObject == 0)
		GetColour(m_aDiffuseColor, Diffuse);

	// set the openGL material properties
	glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,Ambient);
	glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,Diffuse);
	glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,Specular);
	glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,Emission);
	glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,Shininess);


	// if we have normals, enable them
	if(normalCount>0)
	{
		glEnableClientState(GL_NORMAL_ARRAY);
		glNormalPointer( GL_FLOAT, 0, normalArrays[0] );
	}

	// if we have tex coords, enable them
	if(texCoordCount>0)
	{
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer( 2, GL_FLOAT, 0, texCoordArrays[0] );
	}

	// enable the vertex array
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer( 3, GL_FLOAT, 0, vertexArray );

	// draw the vertex arrays using the indices provided
	glDrawElements( prim, indexCount, GL_UNSIGNED_INT, indexArray );


	return MS::kSuccess;
}

// computes the output colour. In a software shader this would be used to calculate an output
// colour for a pixel during rendering. There are 2 occasions when this takes place...
//
// 1. When the scene is rendered
// 2. When the sample swatch in the Attribute Editor gets rendered.
//
// In our case then, it will only get called when the swatch gets evaluated. Only software shaders
// can get rendered by the maya renderer...
//
MStatus BasicHwShader::compute (const MPlug& plug, MDataBlock& data)
{
	// Check that the requested recompute is one of the output values
	//
	if( (plug != outColor) &&
		(plug.parent() != outColor) )
		return MS::kUnknownParameter;

	// get the input colour handle
	MDataHandle inputData = data.inputValue(m_aDiffuseColor);

	// use that to get the color value
	const MFloatVector & color = inputData.asFloatVector();

	// get the outColor handle
	MDataHandle outColorHandle = data.outputValue( outColor );

	// get a reference to the output colour value
    MFloatVector & outColor = outColorHandle.asFloatVector();

	// set the output colour to the input colour
	outColor = color;

	// clean the data block
	data.setClean( plug );

	return MS::kSuccess;
}
