//-------------------------------------------------------------------------------------------------
/// \file	BasicHwShaderBehavior.cpp
/// \date	6-11-04
/// \author	Rob Bateman
/// \brief	Up until now, we have had a couple of problems with our node. For example, if within
/// 		the hypershade we drag the HW shader onto say a lambert or phong, Maya will not make
/// 		a connection to the node.
/// 			Essentially Maya will not know how to connect our nodes to other nodes, so we have
/// 		to provide maya with that information. We do this by simply providing Maya with our own
/// 		class derived from MPxDragAndDropBehavior.
///
/// 			Primarily we have to tell maya when to use this behaviour class by overloading the
///			shouldBeUsedFor function. In addition we can then specify the following functions to
/// 		connect the nodes and attributes correctly....
///
/// 					connectNodeToNode
/// 					connectNodeToAttr
/// 					connectAttrToNode
/// 					connectAttrToAttr
///
//-------------------------------------------------------------------------------------------------

#include "BasicHwShaderBehavior.h"
#include <maya/MIOStream.h>
#include <maya/MGlobal.h>
#include <maya/MPlugArray.h>
#include <maya/MFnDagNode.h>
#include <maya/MObjectArray.h>

//-------------------------------------------------------------------------------------------------
/// \brief	A utility function to connect the attributes
/// \param	srcPlug		-	the input attribute
/// \param	destPlug	-	the output attribute
/// \param	force		-	true if MUST get connected
///
static MStatus connectAttr(const MPlug& srcPlug,const MPlug &destPlug,bool force)
{
	MStatus result = MS::kFailure;

	// make sure both plugs are valid
	if( !srcPlug.isNull() && !destPlug.isNull() )
	{
		// construct a mel command to connect the attributes
		MString cmd = "connectAttr ";

		// if forcing, use the -f flag
		if (force) {
			cmd += "-f ";
		}

		// add the names of the attributes to connect
		cmd += srcPlug.name() + " " + destPlug.name();

		// execute the mel command to connect the attrs
		result = MGlobal::executeCommand(cmd);
	}
	return result;
}

//-------------------------------------------------------------------------------------------------
/// \brief	Used to return a new instance of this behaviour node to Maya
/// \return the new behaviour instance
///
void* BasicHwShaderBehavior::creator()
{
	return new BasicHwShaderBehavior;
}

//-------------------------------------------------------------------------------------------------
/// \brief	We MUST overload this function to tell maya which connections this
/// 		class will handle. ie, does it support connections from one node
/// 		type to another?
/// \param	srcNode		-	the input node
/// \param	dstNode		-	the output node
/// \param	srcPlug		-	the input attribute
/// \param	dstPlug		-	the output attribute
/// \return	true if this behaviour class handles this node connection.
/// \note	Since this class can handle connections between nodes, you may well find
/// 		that the plugs are NULL. To be safe, make sure you test the plugs with
/// 		the member function isNull() before doing anything with them!
///
bool BasicHwShaderBehavior::shouldBeUsedFor( MObject& srcNode,
											 MObject& dstNode,
											 MPlug& srcPlug ,
											 MPlug& dstPlug )
{
	bool result = false;

	MFnDependencyNode fnSrc(srcNode);
	MFnDependencyNode fnDst(dstNode);

	// Handle dropping a hw shader on a maya shader
	if( fnSrc.typeName() == "basicHwShader" && dstNode.hasFn(MFn::kLambert))
	{
		result = true;
	}

	// Handle dropping a file texture onto a hw shader
	if( srcNode.hasFn(MFn::kFileTexture) && fnDst.typeName() == "basicHwShader" )
	{
		result = true;
	}

	return result;
}

//-------------------------------------------------------------------------------------------------
/// \brief	Overide this function to connect nodes to nodes
/// \param	srcNode		-	the input node
/// \param	dstNode		-	the output node
/// \param	force		-	any existing connections should be broken to make
/// 							this connection occur.
///	\return	MS::kSuccess or MS::kFailure
///
MStatus BasicHwShaderBehavior::connectNodeToNode( MObject& srcNode,
												  MObject& dstNode,
												  bool force )
{
	MStatus result = MS::kFailure;

	// attach a function set to the source node
	MFnDependencyNode fnSrc(srcNode);

	// attach a function set to the destination node
	MFnDependencyNode fnDst(dstNode);

	// Handle dropping a hw shader on a maya shader
	if( fnSrc.typeName() == "basicHwShader" && dstNode.hasFn(MFn::kLambert))
	{

		// get the plugs to the destination and source attributes
		MPlug srcPlug = fnSrc.findPlug("outColor");
		MPlug dstPlug = fnDst.findPlug("hardwareShader");

		// connect the 2 attrs
		result = connectAttr( srcPlug, dstPlug, force );
	}

	// Handle dropping a file texture onto a hw shader
	if( srcNode.hasFn(MFn::kFileTexture) && fnDst.typeName() == "basicHwShader" )
	{

		// get the plugs to the destination and source attributes
		MPlug srcPlug = fnSrc.findPlug("outColor");
		MPlug dstPlug = fnDst.findPlug("color");

		// connect the 2 attrs
		result = connectAttr( srcPlug, dstPlug, force );
	}

	return result;
}

//-------------------------------------------------------------------------------------------------
/// \brief	Overide this function to connect nodes to attributes
/// \param	srcNode		-	the input node
/// \param	dstPlug		-	the output attribute
/// \param	force		-	any existing connections should be broken to make
/// 							this connection occur.
///	\return	MS::kSuccess or MS::kFailure
///
MStatus BasicHwShaderBehavior::connectNodeToAttr( MObject& srcNode,
												  MPlug& dstPlug,
												  bool force )

{
	MStatus result = MS::kFailure;

	// attach a function set to the source node
	MFnDependencyNode fnSrc(srcNode);
	// attach a function set to the destination node
	MFnDependencyNode fnDst(dstPlug.node());

	// Handle dropping a hw shader on a maya shader
	if( fnSrc.typeName() == "basicHwShader" && dstPlug.node().hasFn(MFn::kLambert))
	{
		// get the plugs to the destination and source attributes
		MPlug srcPlug = fnSrc.findPlug("outColor");

		// connect the 2 attrs
		result = connectAttr( srcPlug, dstPlug, force );
	}

	// Handle dropping a file texture onto a hw shader
	if( srcNode.hasFn(MFn::kFileTexture) && fnDst.typeName() == "basicHwShader" )
	{

		// get the plugs to the destination and source attributes
		MPlug srcPlug = fnSrc.findPlug("outColor");

		// connect the 2 attrs
		result = connectAttr( srcPlug, dstPlug, force );
	}

	return result;
}

//-------------------------------------------------------------------------------------------------
/// \brief	Overide this function to connect attributes to nodes
/// \param	srcPlug		-	the input attribute
/// \param	dstNode		-	the output node
/// \param	force		-	any existing connections should be broken to make
/// 							this connection occur.
///	\return	MS::kSuccess or MS::kFailure
///
MStatus BasicHwShaderBehavior::connectAttrToNode( MPlug& srcPlug,
												  MObject& dstNode,
												  bool force )
{
	MStatus result = MS::kFailure;

	// get the node of the input connection
	MObject srcNode = srcPlug.node();

	// attach a function set to the source node
	MFnDependencyNode fnSrc(srcNode);

	// attach a function set to the destination node
	MFnDependencyNode fnDst(dstNode);

	// Handle dropping a hw shader on a maya shader
	if( fnSrc.typeName() == "basicHwShader" && dstNode.hasFn(MFn::kLambert))
	{

		// get the plugs to the destination and source attributes
		MPlug dstPlug = fnDst.findPlug("hardwareShader");

		// connect the 2 attrs
		result = connectAttr( srcPlug, dstPlug, force );
	}

	// Handle dropping a file texture onto a hw shader
	if( srcNode.hasFn(MFn::kFileTexture) && fnDst.typeName() == "basicHwShader" )
	{
		// get the plugs to the destination and source attributes
		MPlug dstPlug = fnDst.findPlug("color");

		// connect the 2 attrs
		result = connectAttr( srcPlug, dstPlug, force );
	}

	return result;
}

//-------------------------------------------------------------------------------------------------
/// \brief	Overide this function to connect attributes to attributes
/// \param	srcPlug		-	the input attribute
/// \param	dstPlug		-	the output attribute
/// \param	force		-	any existing connections should be broken to make
/// 							this connection occur.
///	\return	MS::kSuccess or MS::kFailure
///
MStatus BasicHwShaderBehavior::connectAttrToAttr( MPlug& srcPlug,
												  MPlug& dstPlug,
												  bool force )
{
	return connectAttr( srcPlug, dstPlug, force );
}
